import matplotlib.pyplot as plt
import numpy as np
import os
import rdflib
from collections import defaultdict
import pandas as pd

def iguana_results(dataset_dir):
    # rdf/s URIs
    rdf_type_uri = rdflib.URIRef('http://www.w3.org/1999/02/22-rdf-syntax-ns#type')
    rdfs_label_uri = rdflib.URIRef('http://www.w3.org/2000/01/rdf-schema#label')
    rdfs_id = rdflib.URIRef('http://www.w3.org/2000/01/rdf-schema#ID')
    # iguana URIs
    suite_uri = rdflib.URIRef('http://iguana-benchmark.eu/class/Suite')
    query_uri = rdflib.URIRef('http://iguana-benchmark.eu/properties/query')
    query_id_uri = rdflib.URIRef('http://iguana-benchmark.eu/properties/queryID')
    experiment_uri = rdflib.URIRef('http://iguana-benchmark.eu/properties/experiment')
    dataset_uri = rdflib.URIRef('http://iguana-benchmark.eu/properties/dataset')
    task_uri = rdflib.URIRef('http://iguana-benchmark.eu/properties/task')
    connection_uri = rdflib.URIRef('http://iguana-benchmark.eu/properties/connection')
    n_workers_uri = rdflib.URIRef('http://iguana-benchmark.eu/properties/noOfWorkers')
    # metrics' URIs
    task_metrics = {
        'AvgQPS': rdflib.URIRef('http://iguana-benchmark.eu/properties/AvgQPS'),
        'pAvgQPS': rdflib.URIRef('http://iguana-benchmark.eu/properties/penalizedAvgQPS'),
        'QMPH': rdflib.URIRef('http://iguana-benchmark.eu/properties/QMPH')
    }
    query_metrics = {
        'QPS': rdflib.URIRef('http://iguana-benchmark.eu/properties/QPS'),
        'pQPS': rdflib.URIRef('http://iguana-benchmark.eu/properties/penalizedQPS'),
        'failed': rdflib.URIRef('http://iguana-benchmark.eu/properties/failed')
    }
    iguana_rdf_graph = rdflib.Graph()
    for file in sorted(os.listdir(os.path.dirname(dataset_dir))):
        iguana_rdf_graph.parse(os.path.join(dataset_dir, file), format='nt')
    results = dict()
    # iterate over the suites
    for suite in iguana_rdf_graph.subjects(rdf_type_uri, suite_uri):
        # iterate over the experiments
        for experiment in iguana_rdf_graph.objects(suite, experiment_uri):
            # get the dataset of the experiment
            dataset = [d for d in iguana_rdf_graph.objects(experiment, dataset_uri)][0]
            dataset_label = [label.toPython() for label in iguana_rdf_graph.objects(dataset, rdfs_label_uri)][0]
            if dataset_label not in results.keys():
                results[dataset_label] = defaultdict(dict)
            # iterate over the tasks
            for task in sorted(iguana_rdf_graph.objects(experiment, task_uri), key=lambda id: int(str(id).split('/')[-1])):
                n_workers = [x.toPython() for x in iguana_rdf_graph.objects(task, n_workers_uri)][0]
                # iterate over the connections
                for connection in iguana_rdf_graph.objects(task, connection_uri):
                    connection_label = [label.toPython() for label in
                                        iguana_rdf_graph.objects(connection, rdfs_label_uri)][0]
                    if n_workers not in results[dataset_label][connection_label]:
                        results[dataset_label][connection_label][n_workers] = dict()
                        for m in task_metrics.keys():
                            results[dataset_label][connection_label][n_workers][m] = list()
                        for m in query_metrics.keys():
                            results[dataset_label][connection_label][n_workers][m] = list()
                    # collect info for tasks
                    for (tmetric_name, tmetric_uri) in task_metrics.items():
                        results[dataset_label][connection_label][n_workers][tmetric_name].append(
                            [res.toPython() for res in iguana_rdf_graph.objects(task, tmetric_uri)][0]
                        )
                    # collect info for queries
                    for (qmetric_name, qmetric_uri) in query_metrics.items():
                        metric_results = dict()
                        for query in iguana_rdf_graph.objects(task, query_uri):
                            query_id = [res.toPython() for res in iguana_rdf_graph.objects(query, query_id_uri / rdfs_id)][0]
                            query_res = [res.toPython() for res in iguana_rdf_graph.objects(query, qmetric_uri)][0]
                            metric_results[query_id] = query_res
                        results[dataset_label][connection_label][n_workers][qmetric_name].append(metric_results)
    return pd.DataFrame.from_dict(results)


# Generate plots for GraphQL benchmarks
def graphql_plots(tentris_color, tentrisb_color, neo4j_color, neo4jtranslation_color, results_dir):
    colors = dict()
    colors["TentrisGQL"] = tentris_color
    colors["TentrisGQLBase"] = tentrisb_color
    colors["Neo4jC"] = neo4j_color
    colors["Neo4jGQL"] = neo4jtranslation_color
    # iguana results
    graphql_qts_qm_iguana_res = iguana_results(results_dir + "graphql/qm/")
    graphql_all_t_iguana_res = iguana_results(results_dir + "graphql/all/")

    # plot generation for the first benchmark -- pAvgQPS
    graphql_qts_qm_labels = ['QT' + str(i) for i in range(1, 14)]
    graphql_qts_qm_x = np.arange(len(graphql_qts_qm_labels))
    graphql_width = 0.3
    # dataset = graphql_qts_qm_iguana_res.keys()[0]
    tentris_pos = graphql_width
    neo4j_pos = - graphql_width
    metrics = ['pAvgQPS']
    for dataset in graphql_qts_qm_iguana_res.keys():
        graphql_qts_qm_res = graphql_qts_qm_iguana_res[dataset]
        for metric in metrics:
            fig, ax = plt.subplots()
            ax.bar(graphql_qts_qm_x + tentris_pos, graphql_qts_qm_res['TentrisGQL'][1][metric][1::2], graphql_width,
                label='TentrisGQL', log=True, color=tentris_color, edgecolor='white')
            ax.bar(graphql_qts_qm_x,  graphql_qts_qm_res['TentrisGQLBase'][1][metric][1::2], graphql_width,
                label='TentrisGQLBase', log=True, color=tentrisb_color, edgecolor='white', hatch='\/')
            ax.bar(graphql_qts_qm_x + neo4j_pos, graphql_qts_qm_res['Neo4j'][1][metric][1::2], graphql_width,
                label='Neo4jC', log=True, color=neo4j_color, edgecolor='white')
            # cold runs marker
            for tick in graphql_qts_qm_x:
                ax.plot([neo4j_pos+tick-(graphql_width/3), neo4j_pos+tick+(graphql_width/3)], [graphql_qts_qm_res['Neo4j'][1][metric][0::2][tick], graphql_qts_qm_res['Neo4j'][1][metric][0::2][tick]], "k-")
                ax.plot([tick-(graphql_width/3), tick+(graphql_width/3)], [graphql_qts_qm_res['TentrisGQLBase'][1][metric][0::2][tick], graphql_qts_qm_res['TentrisGQLBase'][1][metric][0::2][tick]], "k-")
                ax.plot([tentris_pos+tick-(graphql_width/3), tentris_pos+tick+(graphql_width/3)], [graphql_qts_qm_res['TentrisGQL'][1][metric][0::2][tick], graphql_qts_qm_res['TentrisGQL'][1][metric][0::2][tick]], "k-")
            plt.text(.75, .99, 'LinGBM'+dataset[6:], fontsize=14, ha='left', va='top', transform=ax.transAxes)
            ax.set_ylabel(metric, fontsize=14)
            ax.set_xlabel('Query Templates', fontsize=14)
            ax.set_xticks(graphql_qts_qm_x)
            ax.set_xticklabels(graphql_qts_qm_labels)
            ax.set_xticklabels(graphql_qts_qm_labels, rotation = (360-45), fontsize = 13)
            ax.tick_params(axis='y', which='major', labelsize=14)
            ax.legend(fontsize=12, loc='upper right', ncol=3, columnspacing=0.5, bbox_to_anchor=(1.015, 1.12))
            fig.tight_layout()
            fig.savefig('./plots/' + dataset + '_' + metric.lower() + '.pdf')

        # boxplots for qps -- one boxplot per system
        graphql_qps_width = 0.4
        n_queries = 112
        qps_fig, qps_ax = plt.subplots()
        labels = ["Neo4jC", "TentrisGQLBase", "TentrisGQL"]
        qps_data = dict()
        qps_data["Neo4jC"] = [y for x in graphql_qts_qm_res['Neo4j'][1]['QPS'][1::2] for y in x.values()]
        qps_data["TentrisGQLBase"] = [y for x in graphql_qts_qm_res['TentrisGQLBase'][1]['QPS'][1::2] for y in x.values()]
        qps_data["TentrisGQL"] = [y for x in graphql_qts_qm_res['TentrisGQL'][1]['QPS'][1::2] for y in x.values()]
        step = graphql_qps_width / n_queries
        qps_ax.boxplot(qps_data.values(), widths=graphql_qps_width, showfliers=False, positions=np.arange(len(labels)))
        for i, system in enumerate(labels):
            start = i - 0.2
            end = start + (step * n_queries)
            x = np.linspace(start, end, n_queries)
            qps_ax.scatter(x, qps_data[system], color=colors[system], marker='*')
        plt.text(.01, .99, 'LinGBM'+dataset[6:], fontsize=14, ha='left', va='top', transform=qps_ax.transAxes)
        qps_ax.set_ylabel('QPS', fontsize=14)
        qps_ax.set_yscale('log')
        qps_ax.set_xticks(np.arange(len(labels)))
        qps_ax.set_xticklabels(labels)
        qps_ax.tick_params(axis='both', which='major', labelsize=14)
        qps_fig.tight_layout()
        qps_fig.savefig('./plots/' + dataset + '_qps.pdf')

    # plot generation for second benchmark
    metric = 'pAvgQPS'
    # print(graphql_all_t_iguana_res)
    graphql_all_t_labels = [1, 4, 8, 16]
    for dataset in graphql_qts_qm_iguana_res.keys():
        fig, ax = plt.subplots()
        graphql_all_t_res = graphql_all_t_iguana_res[dataset]

        # values of 4,8,16-workers
        t_values = [w*r for w in sorted(graphql_all_t_res['TentrisGQL'].keys())[1:4] for r in graphql_all_t_res['TentrisGQL'][w][metric]]
        # insert value of 1-worker only
        t_values.insert(0, graphql_all_t_res['TentrisGQL'][1][metric][1])
        ax.plot(graphql_all_t_labels[:4], t_values, label='TentrisGQL', color=tentris_color, marker='*',
                linestyle='dashed')
        ax.text(graphql_all_t_labels[0] - 0.5, t_values[0] + 100, int(t_values[0]), horizontalalignment='left',
                verticalalignment='top', color=tentris_color, fontsize=13)

        for i in range(1, len(t_values)):
            ax.text(graphql_all_t_labels[i], t_values[i], int(t_values[i]), horizontalalignment='right',
                    verticalalignment='bottom', color=tentris_color, fontsize=13)

        # values of 4,8,16-workers
        tb_values = [w*r for w in sorted(graphql_all_t_res['TentrisGQLBase'].keys())[1:4] for r in graphql_all_t_res['TentrisGQLBase'][w][metric]]
        # insert value of 1-worker only
        tb_values.insert(0, graphql_all_t_res['TentrisGQLBase'][1][metric][1])
        ax.plot(graphql_all_t_labels[:4], tb_values, label='TentrisGQLBase', color=tentrisb_color, marker='x', linestyle='dashed')

        for i in range(len(tb_values)-1):
            ax.text(graphql_all_t_labels[i], tb_values[i]+(tb_values[i]*0.15), int(tb_values[i]), horizontalalignment='left',
                    verticalalignment='bottom', color=tentrisb_color, fontsize=13)
        ax.text(graphql_all_t_labels[3]-0.6, tb_values[3]-200, int(tb_values[3]), horizontalalignment='left',
                verticalalignment='bottom', color=tentrisb_color, fontsize=13)
        # values of 4,8,16-workers
        n_values = [w*r for w in sorted(graphql_all_t_res['Neo4j'].keys())[1:4] for r in graphql_all_t_res['Neo4j'][w][metric]]
        # insert value of 1-worker only
        n_values.insert(0, graphql_all_t_res['Neo4j'][1][metric][1])
        ax.plot(graphql_all_t_labels[:4], n_values, label='Neo4j', color=neo4j_color, marker='o', linestyle='dashed')
        for i in range(len(t_values)):
            ax.text(graphql_all_t_labels[i], n_values[i], int(n_values[i]), horizontalalignment='left',
                    verticalalignment='top', color=neo4j_color, fontsize=13)
        plt.text(.01, .99, 'LinGBM'+dataset[6:], fontsize=14, ha='left', va='top', transform=ax.transAxes)
        ax.set_ylabel(metric, fontsize=14)
        ax.set_xlabel('Number of Clients', fontsize=14)
        ax.legend(fontsize=13, loc='lower right')
        ax.set_yscale('log')
        ax.set_xticks(graphql_all_t_labels[:4])
        ax.set_xticklabels(graphql_all_t_labels[:4])
        ax.tick_params(axis='both', which='major', labelsize=14)
        fig.tight_layout()
        fig.savefig('./plots/' + dataset + '_scalability.pdf')

        # histrogram for translation
        transl_fig, transl_ax = plt.subplots()
        transl_ax.bar(0, graphql_all_t_res['Neo4j'][1][metric][1], graphql_width, label='Neo4j', color=neo4j_color, edgecolor='white')
        transl_ax.bar(0.5, graphql_all_t_res['Neo4jTranslation'][1][metric][1], graphql_width, label='Neo4jGQL', color=neo4j_color, edgecolor='white', hatch='\/')
        transl_ax.bar(1, graphql_all_t_res['TentrisGQL'][1][metric][1], graphql_width, label='TentrisGQL', color=tentris_color, edgecolor='white')
        transl_ax.set_xticks([0, 0.5, 1])
        transl_ax.set_ylabel(metric, fontsize=14)
        transl_ax.set_xticklabels(['Neo4j', 'Neo4jGQL', 'TentrisGQL'])
        transl_ax.tick_params(axis='both', which='major', labelsize=14)
        fig.tight_layout()
        transl_fig.tight_layout()
        transl_fig.savefig('./plots/' + dataset + '_translation.pdf')

if __name__ == '__main__':
    result_dir = "./results/"
    # color palette
    colors = list(plt.cm.get_cmap('Dark2').colors)
    # remove colors from the palette
    colors.pop(3)
    tentrisb_color = colors.pop(1)
    tentris_color = colors.pop()
    neo4j_color = colors.pop()
    neo4jtranslation_color = colors.pop()
    graphql_plots(tentris_color, tentrisb_color, neo4j_color, neo4jtranslation_color, result_dir)
